""" Base handler class. """
import asyncio

import ujson as json
from aiohttp import web, multidict
from collections import defaultdict

from muffin.urls import routes_register
from muffin.utils import to_coroutine, abcoroutine


HTTP_METHODS = 'HEAD', 'OPTIONS', 'GET', 'POST', 'PUT', 'PATCH', 'DELETE'


class HandlerMeta(type):

    """ Prepare handlers. """

    _coroutines = set(m.lower() for m in HTTP_METHODS)

    def __new__(mcs, name, bases, params):
        """ Prepare a Handler Class.

        Ensure that the Handler class has a name.
        Ensure that required methods are coroutines.
        Fix the Handler params.

        """
        # Set name
        params['name'] = params.get('name', name.lower())

        # Define new coroutines
        for fname, method in params.items():
            if callable(method) and hasattr(method, '_abcoroutine'):
                mcs._coroutines.add(fname)

        cls = super(HandlerMeta, mcs).__new__(mcs, name, bases, params)

        # Ensure that the class methods are exist and iterable
        if not cls.methods:
            cls.methods = set(method for method in HTTP_METHODS if method.lower() in cls.__dict__)

        elif isinstance(cls.methods, str):
            cls.methods = [cls.methods]

        cls.methods = [method.upper() for method in cls.methods]

        # Ensure that coroutine methods is coroutines
        for name in mcs._coroutines:
            method = getattr(cls, name, None)
            if not method:
                continue
            setattr(cls, name, to_coroutine(method))

        return cls


class DummyApp:

    def __init__(self):
        self.callbacks = defaultdict(list)

    def register(self, *args, handler=None, **kwargs):
        def wrapper(func):
            self.callbacks[handler].append((args, kwargs, func))
            return func
        return wrapper

    def install(self, app, handler):
        for args, kwargs, func in self.callbacks[handler]:
            app.register(*args, handler=handler, **kwargs)(func)
        del self.callbacks[handler]


class Handler(object, metaclass=HandlerMeta):

    """ Handle request. """

    app = DummyApp()
    name = None
    methods = None

    @classmethod
    def from_view(cls, view, *methods, name=None):
        """ Create a handler class from function or coroutine. """
        view = to_coroutine(view)

        if web.hdrs.METH_ANY in methods:
            methods = HTTP_METHODS

        def proxy(self, *args, **kwargs):
            return view(*args, **kwargs)

        params = {m.lower(): proxy for m in methods}
        params['methods'] = methods
        return type(name or view.__name__, (cls,), params)

    @classmethod
    def connect(cls, app, *paths, methods=None, name=None, router=None, view=None):
        """ Connect to the application. """
        if isinstance(cls.app, DummyApp):
            cls.app, dummy = app, cls.app
            dummy.install(app, cls)

        @asyncio.coroutine
        def handler(request):
            return cls().dispatch(request, view=view)

        if not paths:
            paths = ["/%s" % cls.__name__]

        routes_register(
            app, handler, *paths, methods=methods, router=router, name=name or cls.name)

    @classmethod
    def register(cls, *args, **kwargs):
        """ Register view to handler. """
        return cls.app.register(*args, handler=cls, **kwargs)

    @abcoroutine
    def dispatch(self, request, view=None, **kwargs):
        """ Dispatch request. """
        if request.method not in self.methods:
            raise web.HTTPMethodNotAllowed(request.method, self.methods)

        method = getattr(self, view or request.method.lower())
        response = yield from method(request, **kwargs)

        return (yield from self.make_response(request, response))

    @abcoroutine
    def make_response(self, request, response):
        """ Convert a handler result to web response. """

        while asyncio.iscoroutine(response):
            response = yield from response

        if isinstance(response, web.StreamResponse):
            return response

        if isinstance(response, str):
            return web.Response(text=response, content_type='text/html')

        if isinstance(response, (list, dict)):
            return web.Response(text=json.dumps(response), content_type='application/json')

        if isinstance(response, (multidict.MultiDict, multidict.MultiDictProxy)):
            response = dict(response)
            return web.Response(text=json.dumps(response), content_type='application/json')

        if isinstance(response, bytes):
            response = web.Response(body=response, content_type='text/html')
            response.charset = self.app.cfg.ENCODING
            return response

        if response is None:
            response = ''

        return web.Response(text=str(response), content_type='text/html')

    def parse(self, request):
        """ Return a coroutine which parses data from request depends on content-type.

        Usage: ::

            def post(self, request):
                data = yield from self.parse(request)
                # ...

        """
        if request.content_type in ('application/x-www-form-urlencoded', 'multipart/form-data'):
            return request.post()

        if request.content_type == 'application/json':
            return request.json()

        return request.text()



# Configure your tests here
import asyncio
import inspect
import io
import os

import aiohttp
import logging
import pytest
import webob
import webtest
from gunicorn import util

import muffin


def pytest_addoption(parser):
    """ Append pytest options for testing Muffin apps. """
    parser.addini('muffin_app', 'Set path to muffin application')
    parser.addoption('--muffin-app', dest='muffin_app', help='Set to muffin application')

    parser.addini('muffin_config', 'Set module path to muffin configuration')
    parser.addoption('--muffin-config', dest='muffin_config',
                     help='Set module path to muffin configuration')


def pytest_configure(config):
    config.addinivalue_line('markers', 'async: mark test to run asynchronuosly.')


@pytest.mark.tryfirst
def pytest_pycollect_makeitem(collector, name, obj):
    if collector.funcnamefilter(name) and inspect.isgeneratorfunction(obj):
        item = pytest.Function(name, parent=collector)
        if ('async' in item.keywords):
            return list(collector._genfunctions(name, obj))


def pytest_runtest_setup(item):
    if 'async' in item.keywords and 'loop' not in item.fixturenames:
        item.fixturenames.append('loop')


@pytest.mark.tryfirst
def pytest_pyfunc_call(pyfuncitem):
    if 'async' in pyfuncitem.keywords:
        loop = pyfuncitem.funcargs['loop']
        funcargs = pyfuncitem.funcargs
        testargs = {arg: funcargs[arg]
                    for arg in pyfuncitem._fixtureinfo.argnames}
        coro = muffin.to_coroutine(pyfuncitem.obj)
        loop.run_until_complete(asyncio.async(coro(**testargs), loop=loop))
        return True


def pytest_load_initial_conftests(early_config, parser, args):
    """ Prepare to loading Muffin application. """
    options = parser.parse_known_args(args)

    # Initialize configuration
    config = options.muffin_config or early_config.getini('muffin_config')
    if config:
        os.environ[muffin.CONFIGURATION_ENVIRON_VARIABLE] = config

    # Initialize application
    app = options.muffin_app or early_config.getini('muffin_app')
    early_config.app = app


def WSGIHandler(app, loop):

    def handle(environ, start_response):

        req = webob.Request(environ)
        vers = aiohttp.HttpVersion10 if req.http_version == 'HTTP/1.0' else aiohttp.HttpVersion11
        message = aiohttp.RawRequestMessage(
            req.method, req.path_qs, vers, aiohttp.CIMultiDict(req.headers), False, False)
        payload = aiohttp.StreamReader(loop=loop)
        payload.feed_data(req.body)
        payload.feed_eof()
        factory = aiohttp.web.RequestHandlerFactory(
            app, app.router, loop=loop, keep_alive_on=False)
        handler = factory()
        handler.transport = io.BytesIO()
        handler.transport._conn_lost = 0
        handler.transport.get_extra_info = lambda s: ('127.0.0.1', 80)
        handler.writer = aiohttp.parsers.StreamWriter(
            handler.transport, handler, handler.reader, handler._loop)
        coro = handler.handle_request(message, payload)
        if loop.is_running():
            raise RuntimeError('Client cannot running in coroutines')

        loop.run_until_complete(coro)
        handler.transport.seek(0)
        res = webob.Response.from_file(handler.transport)
        start_response(res.status[9:], res.headerlist)
        return res.app_iter

    return handle


@pytest.fixture(scope='session')
def loop(request):
    """ Create and provide asyncio loop. """
    loop = asyncio.get_event_loop()
    asyncio.set_event_loop(loop)
    return loop


@pytest.fixture(scope='session')
def app(pytestconfig, request):
    """ Provide an example application. """
    if pytestconfig.app:
        return util.import_app(pytestconfig.app)

    logging.warn(
        'Improperly configured. Please set ``muffin_app`` in your pytest config. '
        'Or use ``--muffin-app`` command option.')
    return None


@pytest.fixture(scope='session', autouse=True)
def _initialize(app, loop, request):
    app._loop = loop

    if 'peewee' in app.plugins:
        import peewee

        for model in app.plugins.peewee.models.values():
            try:
                model.create_table()
            except peewee.OperationalError:
                pass

    loop.run_until_complete(app.start())

    @request.addfinalizer
    def finish():
        loop.run_until_complete(app.finish())
        loop.close()


@pytest.fixture(scope='function')
def client(app, loop):
    """ Prepare a tests' client. """
    app = WSGIHandler(app, loop=loop)
    client = webtest.TestApp(app)
    client.exception = webtest.AppError
    return client


@pytest.fixture(scope='function')
def db(app, request):
    """ Run tests in transaction. """
    if 'peewee' in app.plugins:
        app.ps.peewee.database.set_autocommit(False)
        app.ps.peewee.database.begin()
        request.addfinalizer(lambda: app.ps.peewee.database.rollback())
        return app.ps.peewee.database


@pytest.fixture(scope='session')
def mixer(app):
    if 'peewee' in app.plugins:
        from mixer.backend.peewee import Mixer
        return Mixer(commit=True)



""" Implement Muffin Application. """
import asyncio
import importlib
import inspect
import logging
import os
import re

from aiohttp import web
from cached_property import cached_property

from muffin import CONFIGURATION_ENVIRON_VARIABLE
from muffin.handler import Handler
from muffin.manage import Manager
from muffin.urls import StaticRoute
from muffin.utils import LStruct, to_coroutine


RETYPE = type(re.compile('@'))


class MuffinException(Exception):

    """ Implement a Muffin's exception. """

    pass


class Application(web.Application):

    """ Improve aiohttp Application. """

    # Default application settings
    defaults = {

        # Path to configuration module
        'CONFIG': None,

        # Enable debug mode
        'DEBUG': False,

        # Default encoding
        'ENCODING': 'utf-8',

        # Logging options
        'LOG_LEVEL': 'WARNING',
        'LOG_FORMAT': '%(asctime)s [%(process)d] [%(levelname)s] %(message)s',
        'LOG_DATE_FORMAT': '[%Y-%m-%d %H:%M:%S %z]',

        # List of plugins
        'PLUGINS': [],

        # Setup static files in development
        'STATIC_PREFIX': '/static',
        'STATIC_FOLDERS': ['static'],
    }

    def __init__(self, name, *, loop=None, router=None, middlewares=(), logger=web.web_logger,
                 handler_factory=web.RequestHandlerFactory, **OPTIONS):
        """ Initialize the application. """
        super(Application, self).__init__(loop=loop, router=router, middlewares=middlewares,
                                          logger=logger, handler_factory=handler_factory)

        self.name = name

        self._error_handlers = {}
        self._start_callbacks = []

        # Overide options
        self.defaults['CONFIG'] = OPTIONS.pop('CONFIG', self.defaults['CONFIG'])
        self.cfg.update(OPTIONS)

        # Setup logging
        ch = logging.StreamHandler()
        ch.setFormatter(logging.Formatter(self.cfg.LOG_FORMAT, self.cfg.LOG_DATE_FORMAT))
        self.logger.addHandler(ch)
        self.logger.setLevel(self.cfg.LOG_LEVEL)
        self.logger.name = 'muffin'

        self.manage = Manager(self)

        # Setup static files option
        if isinstance(self.cfg.STATIC_FOLDERS, str):
            self.cfg.STATIC_FOLDERS = [self.cfg.STATIC_FOLDERS]

        elif not isinstance(self.cfg.STATIC_FOLDERS, list):
            self.cfg.STATIC_FOLDERS = list(self.cfg.STATIC_FOLDERS)

        # Setup plugins
        self.plugins = self.ps = LStruct()
        for plugin in self.cfg.PLUGINS:
            try:
                self.install(plugin)
            except Exception as exc:
                self.logger.error('Plugin is invalid: %s', plugin)
                self.logger.exception(exc)

    def __repr__(self):
        """ Human readable representation. """
        return "<Application: %s>" % self.name

    @cached_property
    def cfg(self):
        """ Load the application configuration. """
        config = LStruct(self.defaults)
        module = config['CONFIG'] = os.environ.get(
            CONFIGURATION_ENVIRON_VARIABLE, config['CONFIG'])

        if module:
            try:
                module = importlib.import_module(module)
                config.update({
                    name: getattr(module, name) for name in dir(module)
                    if name == name.upper() and not name.startswith('_')
                })

            except ImportError:
                config.CONFIG = None
                self.register_on_start(
                    lambda app: app.logger.warn("The configuration hasn't found: %s" % module))

        return config

    def install(self, plugin, name=None):
        """Install a plugin to the application."""
        if isinstance(plugin, str):
            module, _, attr = plugin.partition(':')
            module = importlib.import_module(module)
            plugin = getattr(module, attr or 'Plugin')

        name = name or plugin.name
        if name in self.ps:
            raise MuffinException('Plugin with name `%s` is already intalled.' % name)

        if isinstance(plugin, type):
            plugin = plugin()

        if hasattr(plugin, 'setup'):
            plugin.setup(self)

        if hasattr(plugin, 'middleware_factory') \
                and plugin.middleware_factory not in self.middlewares:
            self.middlewares.append(plugin.middleware_factory)

        if hasattr(plugin, 'start'):
            self.register_on_start(plugin.start)

        if hasattr(plugin, 'finish'):
            self.register_on_finish(plugin.finish)

        # Save plugin links
        self.ps[name] = plugin

    @asyncio.coroutine
    def start(self):
        """Start the application.

        Support for start-callbacks and lock the application's configuration and plugins.

        """
        if self._error_handlers and exc_middleware_factory not in self._middlewares:
            self._middlewares.append(exc_middleware_factory)

        for path in self.cfg.STATIC_FOLDERS:
            if os.path.isdir(path):
                route = StaticRoute(None, self.cfg.STATIC_PREFIX.rstrip('/') + '/', path)
                self.router.register_route(route)

            else:
                self.logger.warn('Disable static folder (hasnt found): %s' % path)

        for (cb, args, kwargs) in self._start_callbacks:
            try:
                res = cb(self, *args, **kwargs)
                if (asyncio.iscoroutine(res) or isinstance(res, asyncio.Future)):
                    yield from res
            except Exception as exc:
                self.loop.call_exception_handler({
                    'message': "Error in start callback",
                    'exception': exc,
                    'application': self,
                })

        # Lock the application's settings and plugin's registry after start
        self.cfg.lock()
        self.ps.lock()

        # Lock plugin's configurations
        for plugin in self.ps.values():
            plugin.cfg.lock()

    def register_on_start(self, func, *args, **kwargs):
        """ Register a start callback. """
        self._start_callbacks.append((func, args, kwargs))

    def register(self, *paths, methods=None, name=None, handler=None):
        """ Register function/coroutine/muffin.Handler on current application. """

        if isinstance(methods, str):
            methods = [methods]

        def wrapper(view):
            if handler is None:

                handler_ = view
                methods_ = methods or [web.hdrs.METH_ANY]

                if inspect.isfunction(handler_) or inspect.ismethod(handler_):
                    handler_ = Handler.from_view(view, *methods_, name=name)

                handler_.connect(self, *paths, methods=methods_, name=name)

            else:

                view_name = view.__name__
                if hasattr(handler, view_name):
                    raise RuntimeError('Handler already have method "%s"' % view_name)
                setattr(handler, view_name, to_coroutine(view))
                name_ = name or view_name
                handler.connect(self, *paths, methods=methods, name=name_, view=view_name)

            return view

        # Support for @app.register(func)
        if len(paths) == 1 and callable(paths[0]):
            view = paths[0]

            if inspect.isclass(view) and issubclass(view, web.HTTPError):
                return wrapper

            paths = []
            return wrapper(view)

        return wrapper


@asyncio.coroutine
def exc_middleware_factory(app, handler):
    """ Handle exceptions. """
    @asyncio.coroutine
    def middleware(request):
        try:
            return (yield from handler(request))
        except Exception as exc:
            if type(exc) in app._error_handlers:
                request.exception = exc
                return (yield from app._error_handlers[type(exc)](request))
            raise
    return middleware


def run():
    """ Run the Gunicorn Application. """
    from .worker import GunicornApp

    GunicornApp("%(prog)s [OPTIONS] [APP]").run()


if __name__ == '__main__':
    run()



import os.path
import re

from aiohttp import web


DYNS_RE = re.compile(r'(\{[^{}]*\})')
DYNR_RE = re.compile(r'^\{(?P<var>[a-zA-Z][_a-zA-Z0-9]*)(?::(?P<re>.+))*\}$')
RETYPE = type(re.compile('@'))


def sre(reg):
    """ Support `Muffin` URL RE. """
    reg = reg.strip('^$')

    def parse(match):
        [part] = match.groups()
        match = DYNR_RE.match(part)
        params = match.groupdict()
        return '(?P<%s>%s)' % (params['var'], params['re'] or '[^{}/]+')

    return re.compile('^%s$' % DYNS_RE.sub(parse, reg))


class RawReRoute(web.DynamicRoute):

    """ Support raw re. """

    def __init__(self, method, handler, name, pattern):
        """ Skip a formatter. """
        super().__init__(method, handler, name, pattern, None)

    def match(self, path):
        match = self._pattern.match(path)
        if match is None:
            return None
        return match.groupdict('')

    def url(self, *, parts, query=None):
        """ Not supported. """
        raise RuntimeError(".url() is not allowed for RawReRoute")

    def __repr__(self):
        """ Fix representation. """
        name = "'" + self.name + "' " if self.name is not None else ""
        return "<RawReRoute {name}[{method}] {pattern} -> {handler!r}".format(
            name=name, method=self.method, pattern=self._pattern, handler=self.handler)


class StaticRoute(web.StaticRoute):

    """ Support multiple static resorces. """

    def match(self, path):
        """ Check for file is exists. """
        if not path.startswith(self._prefix):
            return None

        filename = path[self._prefix_len:]
        filepath = os.path.join(self._directory, filename)
        if not os.path.exists(filepath) or not os.path.isfile(filepath):
            return None

        return {'filename': path[self._prefix_len:]}


def routes_register(app, view, *paths, methods=None, router=None, name=''):
    """ Register routes. """

    if router is None:
        router = app.router

    for method in methods or [web.hdrs.METH_ANY]:
        for path in paths:

            # Register any exception to app
            if isinstance(path, type) and issubclass(path, Exception):
                app._error_handlers[path] = view
                continue

            # Fix route name
            cname, num = name + "-" + method.lower(), 1
            while cname in router:
                cname = name + "-" + str(num)
                num += 1

            # Support regexpa in paths
            if isinstance(path, RETYPE):
                router.register_route(RawReRoute(method.upper(), view, cname, path))
                continue

            # Support custom methods
            method = method.upper()
            if method not in router.METHODS:
                router.METHODS.add(method)

            router.add_route(method, path, view, name=cname)
